(ns clj-wordcloud.word-identifier
  (:require [clojure.string :as s])
  (:import [java.util.regex Pattern]))

(def word-matcher
  "What is a word? It's a sequence of alphabetical characters, for some 
   alphabet (in the case of English, a slightly extended Latin alphabet).
   But some of those characters may be decorated with diacriticals, and
   occasionally characters from other alphabets will be used. In my
   writing, these are encoded as 'HTML entities'.
   
   This regular expression attempts to match anything which is a word.
   It does not have a catalogue of valid HTML entities, so things which have
   the form of HTML entities but which are not valid will be accepted"
  #"([A-Za-z]*&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-fA-F]{1,6});[A-Za-z]*)*|[A-Za-z]*")

(def trimable-matcher
  "Intended to matche a string which contains exactly one word 
   (doesn't work yet)."
  #"^[A-Za-z0-9&;]([A-Za-z]*&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-fA-F]{1,6});[A-Za-z]*)^[A-Za-z0-9&]*")
;
(def candidate-matcher
  "This is ineligant. It identifies a string which MIGHT contain a word, and 
   isolates the bit that might be a word."
  #"[^A-Za-z&;]*(.*)[^A-Za-z&;]*")

(def left-trim
  #"[^A-Za-z&]*(.*)")
(def right-trim
  #"([A-Za-z0-9&;]*).*")

(def whitespace-chars [""       ;;/* dummy empty string for homogeneity */
                       "\\u0009" ;; CHARACTER TABULATION
                       "\\u000A" ;; LINE FEED (LF)
                       "\\u000B" ;; LINE TABULATION
                       "\\u000C" ;; FORM FEED (FF)
                       "\\u000D" ;; CARRIAGE RETURN (CR)
                       "\\u0020" ;; SPACE
                       "\\u0085" ;; NEXT LINE (NEL) 
                       "\\u00A0" ;; NO-BREAK SPACE
                       "\\u1680" ;; OGHAM SPACE MARK
                       "\\u180E" ;; MONGOLIAN VOWEL SEPARATOR
                       "\\u2000" ;; EN QUAD 
                       "\\u2001" ;; EM QUAD 
                       "\\u2002" ;; EN SPACE
                       "\\u2003" ;; EM SPACE
                       "\\u2004" ;; THREE-PER-EM SPACE
                       "\\u2005" ;; FOUR-PER-EM SPACE
                       "\\u2006" ;; SIX-PER-EM SPACE
                       "\\u2007" ;; FIGURE SPACE
                       "\\u2008" ;; PUNCTUATION SPACE
                       "\\u2009" ;; THIN SPACE
                       "\\u200A" ;; HAIR SPACE
                       "\\u2028" ;; LINE SEPARATOR
                       "\\u2029" ;; PARAGRAPH SEPARATOR
                       "\\u202F" ;; NARROW NO-BREAK SPACE
                       "\\u205F" ;; MEDIUM MATHEMATICAL SPACE
                       "\\u3000" ;; IDEOGRAPHIC SPACE 
                       ])

(def whitespace-matcher 
  (Pattern/compile (apply str (concat ["["] whitespace-chars ["]+"]))))

(defn word?
  "True predicate, returning `true` if string `s` matches the pattern
   `word-matcher`, `false` otherwise."
  [s]
  (if (and (string? s)(re-matches word-matcher s)) true false))

(defmacro word??
  "Not a true predicate; is string `s` is a word, returns `s`, else `nil`."
  [s]
  `(when (word? ~s) ~s))

;; (defn word???
;;   [s]
;;   (let [c (re-matches candidate-matcher s)]
;;      (when c (word?? (s/trim (second c))))))

(defn word???
  "When `s` is a string which contains exactly one word, returns that word."
  [s]
  (let [r (re-matches left-trim s)
        c (second (re-matches right-trim (second r)))
        c' (if (and (string? c)
                    (= (last c) \;))
             (subs c 0 (dec (count c)))
             c)
        c'' (when-not (empty? c') c')]
    (word?? c'')))

(defn words
  [s]
  ;; You'd think that this would be as simple as 
  ;; (re-seq word-matcher s), but that doesn't work for me.
  (remove nil?
          (map word??? (s/split s whitespace-matcher))))